{
*****************************************************************************
*  Name    : I2CLCD.BAS                                                     *
*  Author  : Gavin Wiggett (Bitfogav)                                       *
*  Notice  : Copyright (c) 2020                                             *
*          : All Rights Reserved                                            *
*  Date    : 14/06/2020                                                     *
*          :                                                                *
*  Version : 1.0                                                            *
*  Notes   : I2CLCD.bas Lib for use with the PCF8574 controlled LCD.        *
*          : It will work with I2C Hardward or Software Libraries.          *
*          : just include the optonal settings in your main prog.           *
*          :                                                                *
*          : Works with 2line and 4line LCD.                                *
*          :                                                                *
*          : Note: Lib doesn't support reading the PCF8574 or LCD busy flag.*
*          : (this maybe added in a future update).                         *
*****************************************************************************
}

Module I2CLCD

// default module options for software I2C if used - user options can override these values...
#option USE_SI2C = false
#option I2C_SDA = PORTB.2
#option I2C_SCL = PORTB.3

// Use hardware or software I2C, (default is hardware I2C)...
#if IsOption (USE_SI2C) And Not (USE_SI2C in (true, false))
   #error USE_SI2C, "Invalid Option, USE_SI2C must be true or false"
#endIf

// SCL option for software I2C only...
#if IsOption(I2C_SCL) And Not IsValidPortPin(I2C_SCL) 
   #error I2C_SCL, "Invalid option. I2C clock must be a valid port pin."
#endif

// SDA option for software I2C only...
#if IsOption(I2C_SDA) And Not IsValidPortPin(I2C_SDA) 
   #error I2C_SDA, "Invalid option. I2C data line must be a valid port pin."
#endif

// By default, the module includes a call to I2C lcd Initialize() at startup. 
// set this option false to exclude initializer... user MUST then call Initialize()
// you can then change the I2C address by calling Initialize() and adding your 
// desired address for example I2CLCD.Initialize($20).  
#option I2CLCD_INITIALIZE = true
#if Not(I2CLCD_INITIALIZE in (true, false))
   #error I2CLCD_INITIALIZE, "Invalid option"
#endif

// Depending on the #option USE_SI2C setting, will select Hardware or Software I2C... 
#if (_mssp > 0) And Not USE_SI2C
   Include "I2C.bas"
#else
   Include "SI2C.bas"
#endif
   
// public LCD control commands...
Public Const 
   cmdCGRAM           = $40,
   cmdDDRAM           = $60,
   cmdClear           = $01,
   cmdHome            = $02,
   cmdEntryMode       = $04,
   cmdUnderLineON     = $0E,  
   cmdCursorOFF       = $0C,
   cmdCursorON        = $0F,
   cmdBackLightON     = $08,
   cmdBackLightOFF    = $00,
   cmdMoveCursorLeft  = $10,
   cmdMoveCursorRight = $14,
   cmdDisplayON       = $0C,
   cmdDisplayOFF      = $08,
   cmdMoveLeft        = $18,
   cmdMoveRight       = $1E      

// Bring LCD controls into the module...
Private Dim 
   RS,                          // LCD register select pin. 
   I2C_Addrs As Byte,           // I2C device address (default $4E).
   I2CInitialize As Initialize  // Bring Hardware or Software I2C into the module...
Private Dim
   FPosX As Byte = 1,           // local x, y coordinates.
   FPosY As Byte = 1            // initialized as per compiler update 2.2.3.3. 

// Bring LCD backlight into the module...      
Private Dim 
   BackLight_State As Byte = cmdBackLightON   
   
'****************************************************************************
'* Name    : ExpanderWrite                                                  *
'* Purpose : Write data to PCF8574 I2C                                      *
'****************************************************************************
Private Sub ExpanderWrite(pData As Byte) 
   Start()
   WriteByte(I2C_Addrs)
   WriteByte(pData Or BackLight_State)
   Stop() 
End Sub

'****************************************************************************
'* Name    : SetData                                                        *
'* Purpose : Write 4 bit data                 (4 bit interface)             *
'****************************************************************************
Private Sub SetData(pNibble As Byte)   
   pNibble = pNibble Or RS  // Include RS Value To LSB OF Data     
   ExpanderWrite(pNibble Or $04) 
   ExpanderWrite(pNibble And $FB) 
   DelayUS(50) 
End Sub

'****************************************************************************
'* Name    : Command                                                        *
'* Purpose : Write command data to LCD        (4 bit interface)             *
'****************************************************************************
Public Sub Command(pCommand As Byte) 
   RS = 0   
   SetData(pCommand And $F0)
   SetData((pCommand << 4) And $F0)
   DelayUS(50)   
End Sub

'****************************************************************************
'* Name    : WriteItem (OVERLOAD)                                           *
'* Purpose : Write a single byte to the LCD   (4 bit interface)             *
'****************************************************************************
Private Sub WriteItem(pData As Byte)
   RS = 1
   SetData(pData And $F0)
   SetData((pData << 4) And $F0)
   DelayUS(50)   
End Sub

'****************************************************************************
'* Name    : WriteItem (OVERLOAD)                                           *
'* Purpose : Write a string to the LCD                                      *
'****************************************************************************
Private Sub WriteItem(pText As String)
   Dim TextPtr As POSTINC0
   Dim Text As INDF0
   FSR0 = AddressOf(pText)
   While Text <> 0
      WriteItem(TextPtr)
   Wend
End Sub

'****************************************************************************
'* Name    : WriteItem (OVERLOAD)                                           *
'* Purpose : Initialise the CGRAM with constant data array. Eight           *
'*         : programmable characters are available (0..7).                  *
'****************************************************************************
Private Sub WriteItem(ByRefConst pBitmap() As Byte)
   Dim Index As Byte
   Command(cmdCGRAM)
   For Index = 0 To Bound(pBitmap)
      WriteItem(pBitmap(Index))              
   Next
   Command(cmdDDRAM)
End Sub

'****************************************************************************
'* Name    : Write (COMPOUND)                                               *
'* Purpose : Calls one or more WriteItem() subroutines                      *
'****************************************************************************
Public Compound Sub Write(WriteItem)

'****************************************************************************
'* Name    : MoveCursor                                                     *
'* Purpose : Move the cursor to line and column                             *
'****************************************************************************
Public Sub MoveCursor(pLine, pCol As Byte) 
   Dec(pCol)   
   Select pLine
      Case 1 : Command($80 + pCol)
      Case 2 : Command($C0 + pCol)
      Case 3 : Command($94 + pCol)
      Case 4 : Command($D4 + pCol)  
   EndSelect 
End Sub

'****************************************************************************
'* Name    : MoveTo (PRIVATE)                                               *
'* Purpose : Moves cursor to module private x, y location                   *
'****************************************************************************
Private Inline Sub MoveTo()
   MoveCursor(FPosY, FPosX)
End Sub 

'****************************************************************************
'* Name    : SetLocationX (PRIVATE)                                         *
'* Purpose : Set cursor x                                                   *
'**************************************************************************** 
Private Inline Sub SetLocationX(pX As FPosX)
End Sub

'****************************************************************************
'* Name    : SetLocationY (PRIVATE)                                         *
'* Purpose : Set cursor y                                                   *
'****************************************************************************
Private Inline Sub SetLocationY(pY As FPosY)
End Sub

'****************************************************************************
'* Name    : WriteAt (COMPOUND)                                             *
'* Purpose : Calls one or more WriteItem() subroutines at location x, y     *
'****************************************************************************
Public Compound Sub WriteAt(SetLocationY, SetLocationX, MoveTo, WriteItem) 

'****************************************************************************
'* Name    : Cls                                                            *
'* Purpose : Clear the LCD screen                                           *
'****************************************************************************
Public Sub Cls()
   Command($01)  
   DelayMS(30) 
End Sub

'****************************************************************************
'* Name    : Backlight                                                      *
'* Purpose : Set LCD backlight on/off                                       *
'****************************************************************************
Public Sub Backlight(pSet As Boolean = true)
   If pSet = true Then
      BackLight_State = cmdBackLightON   
   Else 
      BackLight_State = cmdBackLightOFF
   End If 
   ExpanderWrite($00) 
End Sub

'****************************************************************************
'* Name    : ShiftLeft                                                      *
'* Purpose : Shift the LCD display left                                     *
'****************************************************************************
Public Sub ShiftLeft()
   Command($18)    
   DelayUS(40) 
End Sub

'****************************************************************************
'* Name    : ShiftRight                                                     *
'* Purpose : Shift the LCD display right                                    *
'****************************************************************************
Public Sub ShiftRight()
   Command($1C) 
   DelayUS(40)    
End Sub

'****************************************************************************
'* Name    : Initialize (Public)                                            *
'* Purpose : Initialize the LCD - Optional User can define i2c device addr  *
'****************************************************************************
Public Sub Initialize(pAddr As Byte = $4E)
   DelayMS(100)   // wait for LCD to stabilise... 
   I2C_Addrs = pAddr
   ExpanderWrite($00) 
   DelayMS(30) 
   Command($03) 
   DelayMS(5) 
   Command($03) 
   DelayMS(5) 
   Command($03) 
   DelayMS(5) 
   Command(cmdHome) 
   DelayMS(5) 
   Command($20 Or (2 << 2))  ' function set (4-bit interface, 2 lines, 5*7 Pixels) 
   DelayMS(50) 
   Command(cmdDisplayON) 
   DelayMS(50) 
   Command(cmdClear) 
   DelayMS(50) 
   Command(cmdEntryMode Or cmdHome) 
   DelayMS(50) 
End Sub

// initialise the I2C module (HW or SW)...  
I2CInitialize()

// initialise the I2CLCD module, you must disable the option above
// if you wish to use your own I2C address, see the option above...
#if I2CLCD_INITIALIZE
Initialize()
#endif          

End Module

